@@ -1,31 +1,13 @@
-import grpc
 import google.cloud.datastore.helpers as ds_helpers
 from google.cloud import ndb
 from google.cloud.datastore_v1 import types
 from google.cloud.datastore_v1.proto import datastore_pb2_grpc
-from typing import Dict, List, Optional, NamedTuple
+from typing import List
 
-from .futures import InstantFuture
-
-
-class _StoredObject(NamedTuple):
-    version: int
-    entity: types.Entity
-
-
-class _RequestWrapper(grpc.UnaryUnaryMultiCallable):
-    def __init__(self, func):
-        self.func = func
-
-    def __call__(self, request, *args, **kwargs):
-        return self.func(request, *args, **kwargs)
-
-    def with_call(self, request, *args, **kwargs):
-        return self(request, *args, **kwargs)
-
-    def future(self, request, *args, **kwargs):
-        resp = self(request, *args, **kwargs)
-        return InstantFuture(resp)
+from ._in_memory_store import _InMemoryStore
+from ._request_wrapper import _RequestWrapper
+from ._stored_object import _StoredObject
+from ._transactions import _TransactionType
 
 
 class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
@@ -39,41 +21,40 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
         # TODO missing HAS_ANCESTOR
     }
 
-    seqid: int
-    store: Dict[str, _StoredObject]
+    store: _InMemoryStore
 
     Lookup: _RequestWrapper
     Commit: _RequestWrapper
     RunQuery: _RequestWrapper
-    # BeginTransaction: _RequestWrapper
-    # Rollback: _RequestWrapper
+    BeginTransaction: _RequestWrapper
+    Rollback: _RequestWrapper
     # AllocateIds: _RequestWrapper
     # ReserveIds: _RequestWrapper
 
     def __init__(self) -> None:
-        # don't call super, we want to do other stuff
-        # also we can ignore the channel, since we're doing it all in memory
-        self.seqid = 0
-        self.store = {}
+        self.store = _InMemoryStore()
 
         self.Lookup = _RequestWrapper(self._lookup)
         self.Commit = _RequestWrapper(self._commit)
         self.RunQuery = _RequestWrapper(self._run_query)
+        self.BeginTransaction = _RequestWrapper(self._begin_transaction)
+        self.Rollback = _RequestWrapper(self._rollback)
 
     def _insert_model(self, model: ndb.Model) -> None:
         ds_key = model.key._key.to_protobuf()
-        assert self._get_stored_data(ds_key) is None
+        assert self.store.get(ds_key, None) is None
         entity_proto = ndb.model._entity_to_protobuf(model)
-        self._put_entity(entity_proto, 0)
+        self.store.put(entity_proto, 0, None)
 
     def _lookup(
         self, request: types.LookupRequest, *args, **kwargs
     ) -> types.LookupResponse:
         found: List[types.EntityResult] = []
         missing: List[types.EntityResult] = []
+        transaction_id = request.read_options.transaction
 
         for key in request.keys:
-            stored_data = self._get_stored_data(key)
+            stored_data = self.store.get(key, transaction_id)
             if stored_data:
                 found.append(
                     types.EntityResult(
@@ -83,21 +64,41 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
             else:
                 missing.append(
                     types.EntityResult(
-                        entity=types.Entity(key=key), version=self.seqid,
+                        entity=types.Entity(key=key),
+                        version=self.store.seqid(transaction_id),
                     )
                 )
 
         return types.LookupResponse(found=found, missing=missing,)
 
+    def _begin_transaction(
+        self, request: types.BeginTransactionRequest, *args, **kwargs
+    ) -> types.BeginTransactionResponse:
+        transaction_mode = None
+        request_type = request.transaction_options.WhichOneof("mode")
+        if request_type == "read_write":
+            transaction_mode = _TransactionType.READ_WRITE
+        elif request_type == "read_only":
+            transaction_mode = _TransactionType.READ_ONLY
+        assert transaction_mode is not None
+        transaction_id = self.store.beginTransaction(transaction_mode)
+
+        return types.BeginTransactionResponse(transaction=transaction_id,)
+
     def _commit(
         self, request: types.CommitRequest, *args, **kwargs
     ) -> types.CommitResponse:
-        results: List[types.MutationResult] = []
-        for mutation in request.mutations:
-            result = self._apply_mutation(mutation)
-            results.append(result)
+        results: List[types.MutationResult] = self.store.commitTransaction(
+            request.transaction, request.mutations
+        )
         return types.CommitResponse(mutation_results=results, index_updates=0,)
 
+    def _rollback(
+        self, request: types.RollbackRequest, *args, **kwargs
+    ) -> types.RollbackResponse:
+        self.store.rollbackTransaction(request.transaction)
+        return types.RollbackResponse()
+
     def _run_query(
         self, request: types.RunQueryRequest, *args, **kwargs
     ) -> types.RunQueryResponse:
@@ -107,8 +108,9 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
 
         # Query processing will be very naive.
         query: types.Query = request.query
+        transaction_id: bytes = request.read_options.transaction
         resp_data: List[_StoredObject] = []
-        for _, stored in self.store.items():
+        for _, stored in self.store.items(transaction_id):
             if query.kind and stored.entity.key.path[-1].kind != query.kind[0].name:
                 # this doesn't account for ancestor keys
                 continue
@@ -131,7 +133,7 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
                 reverse=order.direction == types.PropertyOrder.Direction.DESCENDING,
             )
 
-        if query.limit:
+        if query.HasField("limit"):
             resp_data = resp_data[: query.limit.value]
 
         return types.RunQueryResponse(
@@ -141,28 +143,18 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
                     types.EntityResult(entity=resp.entity, version=resp.version,)
                     for resp in resp_data
                 ],
-                snapshot_version=self.seqid,
+                snapshot_version=self.store.seqid(transaction_id),
             )
         )
 
-    def _put_entity(self, ds_entity: types.Entity, entity_version: int) -> None:
-        self.seqid += 1
-        key_str = ds_entity.key.SerializeToString()
-        self.store[key_str] = _StoredObject(entity=ds_entity, version=entity_version)
-
-    def _get_stored_data(self, key: types.Key) -> Optional[_StoredObject]:
-        key_str = key.SerializeToString()
-        return self.store.get(key_str)
-
-    def _delete_entity(self, key: types.Key) -> None:
-        key_str = key.SerializeToString()
-        if key_str in self.store:
-            del self.store[key_str]
-
     def _matches_filter(
         self, stored_obj: _StoredObject, query_filter: types.Filter
     ) -> bool:
         filter_type = query_filter.WhichOneof("filter_type")
+        if filter_type is None:
+            # If doing a query for all entities, filter will be None
+            return True
+
         assert filter_type in ["property_filter", "composite_filter"]
         if filter_type == "property_filter":
             return self._matches_property_filter(
@@ -193,48 +185,3 @@ class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):
                 assert method_name
                 return getattr(prop_val, method_name)(filter_val)
         return False
-
-    def _apply_mutation(self, mutation: types.Mutation) -> types.MutationResult:
-        # TODO will need to potentially do key assignment for insert/upsert
-        mutation_key = self._mutation_key(mutation)
-        existing_data = self._get_stored_data(mutation_key)
-        if mutation.insert:
-            # TODO figure out how to properly express error handling
-            assert existing_data is None
-            self._put_entity(mutation.insert, 0)
-            return types.MutationResult(key=mutation_key, version=0)
-        elif mutation.update:
-            # TODO figure out better error handling
-            assert existing_data is not None
-            if existing_data.version != mutation.base_version:
-                return self._mutation_conflict(mutation_key, existing_data.version)
-            new_version = existing_data.version + 1
-            self._put_entity(mutation.upsert, new_version)
-            return types.MutationResult(key=mutation_key, version=new_version)
-        elif mutation.upsert:
-            if existing_data and existing_data.version != mutation.base_version:
-                return self._mutation_conflict(mutation_key, existing_data.version)
-            new_version = existing_data.version + 1 if existing_data else 0
-            self._put_entity(mutation.upsert, new_version)
-            return types.MutationResult(key=mutation_key, version=new_version)
-        elif mutation.delete:
-            new_version = existing_data.version + 1 if existing_data else 0
-            self._delete_entity(mutation_key)
-            return types.MutationResult(key=mutation_key, version=new_version)
-
-    def _mutation_key(self, mutation: types.Mutation) -> types.Key:
-        if mutation.insert:
-            return mutation.insert.key
-        elif mutation.update:
-            return mutation.update.key
-        elif mutation.upsert:
-            return mutation.upsert.key
-        elif mutation.delete:
-            return mutation.delete.key
-
-    def _mutation_conflict(
-        self, key: types.Key, old_version: int
-    ) -> types.MutationResult:
-        return types.MutationResult(
-            key=key, version=old_version, conflict_detected=True,
-        )
